import BigNumber from "bignumber.js";
declare const DecimalExponent = "DecimalExponent";
declare const BinarySI = "BinarySI";
declare const DecimalSI = "DecimalSI";
/**
 * bePair contains a base and exponent pair
 */
declare class bePair {
    private base;
    private exponent;
    constructor(base: number, exponent: number);
}
/**
 * listSuffixer contains two mapping relationship
 * @class listSuffixer
 */
declare class listSuffixer {
    private suffixToBE;
    private beToSuffix;
    /**
     * @constructor listSuffixer
     */
    constructor();
    /**
     * register a suffix and bePair mapping relationship
     *
     * @param {string} suffix
     * @param {bePair} bePair
     */
    addSuffix(suffix: string, bePair: bePair): void;
    /**
     * Find the bePair related to the given suffix,
     * if not, if boolean in returns will be false
     *
     * @param {string} suffix
     * @returns {[number, number, boolean]} if suffix is not registered, the return boolean will be false
     */
    lookup(suffix: string): [number, number, boolean];
    /**
     * construct a suffix related to the given bePair,
     * if not, the boolean in returns will be false
     *
     * @param {number} base
     * @param {number} exponent
     * @returns {[string, boolean]}
     */
    construct(base: number, exponent: number): [string, boolean];
}
/**
 * suffixHandler help you to handle suffix and format
 * @class suffixHandler
 */
declare class suffixHandler {
    decSuffixes: listSuffixer;
    binSuffixes: listSuffixer;
    /**
     * @constructor suffixHandler
     */
    constructor();
    /**
     * interpret a suffix to base, exponent, format
     *
     * @param {string} suffix
     * @returns {[number, number, string, boolean]}
     */
    interpret(suffix: string): [number, number, string, boolean];
    /**
     * construct a suffix by given base, exponent and format
     *
     * @param {number} base
     * @param {number} exponent
     * @param {string} format
     * @returns {[string, boolean]}
     */
    construct(base: number, exponent: number, format: string): [string, boolean];
}
declare class fastLookup extends suffixHandler {
    constructor();
    /**
     * interpret a suffix to base, exponent, format
     *
     * @param {string} suffix
     * @returns {[number, number, string, boolean]}
     */
    interpret(suffix: string): [number, number, string, boolean];
}
/**
 * quantitySuffixer handles suffixes for all three formats that quantity can handle.
 */
declare const quantitySuffixer: fastLookup;
/**
 * Quantity extends BigNumber, rewrite the string function
 * @class Quantity
 */
declare class Quantity extends BigNumber {
    format: string;
    private digit;
    private suffix;
    /**
     *
     * @param {string} format must be DecimalExponent | DecimalSI | BinarySI
     * @param {number|string|BigNumber|Quantity} numberlike
     * @param {number} base
     */
    constructor(format: string, numberlike: number | string | BigNumber, base?: number);
    sign(): number;
    neg(): any;
    minus(n: number | string | BigNumber, base?: number): any;
    sub(n: number | string | BigNumber, base?: number): any;
    add(n: number | string | BigNumber, base?: number): any;
    plus(n: number | string | BigNumber, base?: number): any;
    div(n: number | string | BigNumber, base?: number): any;
    mul(n: number | string | BigNumber, base?: number): any;
    times(n: number | string | BigNumber, base?: number): any;
    mod(n: number | string | BigNumber, base?: number): any;
    modulo(n: number | string | BigNumber, base?: number): any;
    round(dp?: number, rm?: number): any;
    ceil(): any;
    floor(): any;
    toPrecision(sd?: number, rm?: number): any;
    /**
     * Returns a string representing the value of this BigNumber in the base `10`
     */
    toString(): string;
    /**
     * Convert this quantity to a quantity with specific digit and suffix.
     * The digit is a BigNumber rounded by rounding mode `rm` to a maximum of `dp` decimal places
     *
     * @param {string} suffix
     * @param {number} [dp] If `dp` is omitted, or is `null` or `undefined`, the return quantity.digit is rounded to a whole number.
     * @param {number} [rm] If `rm` is omitted, or is `null` or `undefined`, `ROUNDING_MODE` is used.
     */
    convertTo(suffix: string, dp?: number, rm?: number): any;
}
/**
 * parse string to Quantity
 *
 * @param {string} str
 * @throws {ErrFormatWrong} if quantity format is not match "^([+-]?[0-9.]+)([eEinumkKMGTP]*[-+]?[0-9]*)$"
 * @throws {ErrSuffix} if suffix is not registered
 * @returns {Quantity}
 */
declare function parseQuantity(str: string): Quantity;
export { DecimalExponent, DecimalSI, BinarySI, parseQuantity, quantitySuffixer, Quantity };
